from randog.exceptions import FactoryConstructionError
from ._construct import factory as factory_from_model, factory_from_column


def custom(example, **kwargs):
    """custom_func for `randog.factory.from_example` to generate sqlalchemy-derived objects

    This function is an add-on for `randog.factory.from_example`.

    Specify this function as `custom_func` if you want to create a factory that generates sqlalchemy-derived objects in
    `from_example`.

    Examples
    --------
    >>> from sqlalchemy import Column, Integer, String
    >>> from sqlalchemy.orm import declarative_base
    >>> import randog.factory
    >>> import randog.sqlalchemy
    >>>
    >>> Base = declarative_base()
    >>>
    >>> class User(Base):
    ...     __tablename__ = "user"
    ...
    ...     id = Column(Integer, primary_key=True, autoincrement=True)
    ...     name = Column(String)
    >>>
    >>> # specify `randog.sqlalchemy.custom` as `custom_func`
    >>> factory = randog.factory.from_example(User, custom_func=randog.sqlalchemy.custom)
    >>> generated = factory.next()
    """
    if kwargs.get("context") is not None:
        rnd = kwargs["context"].rnd
    else:
        rnd = kwargs.get("rnd")

    try:
        factory = factory_from_model(example, rnd=rnd)
    except FactoryConstructionError:
        factory = None

    if factory is None:
        try:
            factory = factory_from_column(example, rnd=rnd)
        except FactoryConstructionError:
            factory = None

    if factory is None:
        return NotImplemented
    else:
        return factory



import typing as t
from random import Random

import rstr

from .._base import Factory


class StrRegexRandomFactory(Factory[str]):
    """factory generating random str values"""

    _regex: str

    def __init__(
        self,
        *,
        regex: t.Union[str, t.Pattern],
        rnd: t.Optional[Random] = None,
    ):
        """Return a factory generating random str values.

        Parameters
        ----------
        regex : str
            regular expression for generated string
        rnd : Random, optional
            random number generator to be used

        Raises
        ------
        FactoryConstructionError
            When the specified generating conditions are inconsistent.
        """
        self._regex = regex

    def _next(self) -> str:
        return rstr.xeger(self._regex)



import math
from random import Random
import typing as t

from ._base import Factory
from ._by_iterator import by_iterator
from ..exceptions import FactoryConstructionError


def increment(
    initial_value: t.Optional[int] = None,
    maximum: t.Optional[int] = None,
    *,
    rnd: t.Optional[Random] = None,
) -> Factory[t.Any]:
    """Return a factory which returns sequential numbers.

    Parameters
    ----------
    initial_value : int, optional
        the first value
    maximum : int, optional
        the maximum value. If the generated value reaches the maximum value, 1 is generated next.
        If the maximum value is not specified, it is not reset to 1.
    rnd : Random, optional
        It is not normally used, but it can be accepted as an argument to match other Factory construction functions.

    Raises
    ------
    FactoryConstructionError
        if it is not satisfied `1 <= initial_value <= maximum`
    """
    if initial_value is None:
        initial_value = 1
    if maximum is None:
        maximum = math.inf

    if not (1 <= initial_value <= maximum):
        raise FactoryConstructionError(
            "arguments of increment(initial_value, maximum) must satisfy 1 <= initial_value <= maximum"
        )

    return by_iterator(_increment(initial_value, maximum))


def _increment(initial_value: int, maximum: int) -> t.Iterator[int]:
    next_value = initial_value
    while True:
        yield next_value
        next_value += 1

        if next_value > maximum:
            next_value = 1



import csv
import json
import random
import sys
import typing as t

import randog.factory
from .factory import FactoryDef, FactoryStopException
from ._main import Args, Subcmd, get_subcmd_def


def _build_factories(args: Args) -> t.Iterator[randog.factory.Factory]:
    subcmd_def = get_subcmd_def(args.sub_cmd)

    if args.sub_cmd == Subcmd.Byfile:
        for filepath in args.factories:
            if filepath == "-":
                factory_def = randog.factory.from_pyfile(sys.stdin, full_response=True)
            else:
                factory_def = randog.factory.from_pyfile(filepath, full_response=True)
            factory = factory_def.factory

            post_process = _gen_post_function(args, factory_def)
            if post_process is not None:
                factory = factory.post_process(post_process)

            yield factory
    else:
        iargs, kwargs = subcmd_def.build_args(args)
        yield subcmd_def.get_factory_constructor()(*iargs, **kwargs)


def _get_csv_field(pre_value: t.Mapping, col) -> t.Any:
    if isinstance(col, t.Callable):
        return col(pre_value)
    else:
        return pre_value.get(col)


def _gen_post_function(
    args: Args, factory_def: FactoryDef
) -> t.Optional[t.Callable[[t.Any], t.Any]]:
    """args に応じて、factory_def.factory に適用する post_process 関数を作成する。"""

    # CSV 出力の場合、CSVの行 (Sequence[Any]) を生成するような post_process を作成する。
    if args.get("csv", False):
        if factory_def.csv_columns is not None:

            def _post_function(pre_value) -> t.Optional[t.Sequence[t.Any]]:
                if pre_value is None:
                    return None
                elif isinstance(pre_value, t.Mapping):
                    return [
                        _get_csv_field(pre_value, col)
                        for col in factory_def.csv_columns
                    ]
                elif isinstance(pre_value, t.Sequence):
                    return pre_value
                else:
                    # TODO: 警告
                    return [pre_value]

        else:

            def _post_function(pre_value) -> t.Optional[t.Sequence[t.Any]]:
                if pre_value is None:
                    return None
                elif isinstance(pre_value, t.Mapping):
                    # TODO: 警告
                    return list(pre_value.values())
                elif isinstance(pre_value, t.Sequence) and not isinstance(
                    pre_value, str
                ):
                    return pre_value
                else:
                    # TODO: 警告
                    return [pre_value]

        return _post_function
    else:
        return None


class _DummyIO:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def _open_output_fp_only(args: Args) -> t.Union[_DummyIO, t.TextIO]:
    if args.output_path is None or args.multiple_output_path:
        return _DummyIO()
    else:
        options = {
            "mode": "wt",
        }
        if args.get("csv", False):
            options["newline"] = ""
        return open(args.output_path, **options)


def _open_output_fp_numbered(args: Args, number: int) -> t.Union[_DummyIO, t.TextIO]:
    if args.output_path is None or not args.multiple_output_path:
        return _DummyIO()
    else:
        options = {
            "mode": "wt",
        }
        if args.get("csv", False):
            options["newline"] = ""
        return open(args.output_path_for(number), **options)


def _output_generated(generated: t.Any, fp: t.TextIO, args: Args):
    if args.output_fmt == "repr":
        print(repr(generated), file=fp)
    elif args.output_fmt == "json":
        json.dump(generated, fp, default=_json_default(args))
        fp.write("\n")
    else:
        if args.iso and hasattr(generated, "isoformat"):
            print(generated.isoformat(), file=fp)
        elif args.format:
            print(generated.__format__(args.format), file=fp)
        else:
            print(generated, file=fp)


def _json_default(args: Args):
    if args.iso:
        return lambda v: v.isoformat()
    elif args.format:
        return lambda v: v.__format__(args.format)
    else:
        return str


def _output_to_csv(
    factory: randog.factory.Factory[t.Optional[t.Sequence[t.Sequence[t.Any]]]],
    line_num: int,
    fp: t.TextIO,
    regenerate: float,
    discard: float,
    raise_on_factory_stopped: bool,
):
    csv_writer = csv.writer(fp, lineterminator="\n")
    csv_writer.writerows(
        filter(
            lambda x: x is not None,
            factory.iter(
                line_num,
                regenerate=regenerate,
                discard=discard,
                raise_on_factory_stopped=raise_on_factory_stopped,
            ),
        )
    )


def main():
    args = Args(sys.argv)

    try:
        with _open_output_fp_only(args) as fp_only:
            index = 0
            for factory in _build_factories(args):
                for r_index in range(args.repeat):
                    with _open_output_fp_numbered(args, index) as fp_numbered:
                        index += 1
                        # args と index に応じて出力先 fp を決定する。
                        # args と index に応じて fp_numbered, fp_only の状態が異なるので、それを条件に使用する。
                        if not isinstance(fp_numbered, _DummyIO):
                            fp = fp_numbered
                        elif not isinstance(fp_only, _DummyIO):
                            fp = fp_only
                        else:
                            fp = sys.stdout

                        # 生成処理と出力処理
                        # CSV 出力の場合に生成の方法が異なるので、生成と出力をひとまとめにした。
                        if args.csv is not None:
                            _output_to_csv(
                                factory,
                                args.csv,
                                fp,
                                regenerate=args.regenerate,
                                discard=args.discard,
                                raise_on_factory_stopped=args.error_on_factory_stopped,
                            )
                        else:
                            if args.list is not None:
                                generated = list(
                                    factory.iter(
                                        args.list,
                                        regenerate=args.regenerate,
                                        discard=args.discard,
                                        raise_on_factory_stopped=args.error_on_factory_stopped,
                                    )
                                )
                            else:
                                try:
                                    while True:
                                        generated = factory.next(
                                            raise_on_factory_stopped=args.error_on_factory_stopped
                                        )
                                        if random.random() >= args.regenerate:
                                            break
                                except StopIteration:
                                    break
                                if random.random() < args.discard:
                                    continue

                            _output_generated(generated, fp, args=args)
    except FactoryStopException:
        print(
            "error: the factory stopped generating before the process was complete",
            file=sys.stderr,
        )
        exit(1)


if __name__ == "__main__":
    main()



import argparse
import typing as t

import randog.factory
from ..._utils.type import time
from .. import Args, Subcmd
from ._base import SubcmdDef, add_common_arguments


class SubcmdDefTime(SubcmdDef):
    def cmd(self) -> Subcmd:
        return Subcmd.Time

    def add_parser(self, subparsers) -> argparse.ArgumentParser:
        time_parser = subparsers.add_parser(
            Subcmd.Time.value,
            usage="python -m randog time [MINIMUM MAXIMUM] [--iso | --fmt FORMAT] [common-options]",
            description="It generates values of type datetime.time.",
            add_help=False,
        )
        time_args_group = time_parser.add_argument_group("arguments")
        time_args_group.add_argument(
            "minimum",
            type=time,
            nargs="?",
            metavar="MINIMUM",
            help="the minimum value with the ISO-8601 format. "
            "If not specified, the behavior is left to the specification of randog.factory.randtime.",
        )
        time_args_group.add_argument(
            "maximum",
            type=time,
            nargs="?",
            metavar="MAXIMUM",
            help="the maximum value with the ISO-8601 format. "
            "If not specified, the behavior is left to the specification of randog.factory.randtime.",
        )
        group_date_fmt = time_args_group.add_mutually_exclusive_group()
        group_date_fmt.add_argument(
            "--iso",
            action="store_true",
            help="if specified, it outputs generated object with ISO-8601 format"
            "Since the str of datetime.time is identical to the iso-8601 format, this option is meaningless.",
        )
        group_date_fmt.add_argument(
            "--fmt",
            dest="format",
            metavar="FORMAT",
            help="if specified, it outputs generated object with the specified format, such as '%%H:%%M:%%S.%%f'",
        )
        add_common_arguments(time_parser)

        return time_parser

    def validate_parser(self, args: Args, subparser: argparse.ArgumentParser):
        if args.sub_cmd != Subcmd.Time:
            return

        if args.output_fmt == "repr" and args.iso:
            subparser.error("argument --iso: not allowed with argument --repr")
        elif args.output_fmt == "repr" and args.format:
            subparser.error("argument --fmt: not allowed with argument --repr")

    def build_args(
        self, args: Args
    ) -> t.Tuple[t.Sequence[t.Any], t.Mapping[str, t.Any]]:
        return (args.get("minimum"), args.get("maximum")), {}

    def get_factory_constructor(self) -> t.Callable:
        return randog.factory.randtime



import argparse
import typing as t
from fractions import Fraction

from .. import Args, Subcmd
from ._base import SubcmdDef, add_common_arguments
from ..._utils.type import positive_int, probability
from ...factory import REGENERATE_PROB_MAX


class SubcmdDefByfile(SubcmdDef):
    def cmd(self) -> Subcmd:
        return Subcmd.Byfile

    def add_parser(self, subparsers) -> argparse.ArgumentParser:
        byfile_parser = subparsers.add_parser(
            Subcmd.Byfile.value,
            usage=(
                "python -m randog byfile FACTORY_PATH [FACTORY_PATH ...] [--regenerate PROB_REGEN] "
                "[--discard PROB_DISCARD] [--csv ROW_NUM] [--error-on-factory-stopped] [common-options]"
            ),
            description="It generates values according to factory definition files.",
            add_help=False,
        )
        byfile_args_group = byfile_parser.add_argument_group("arguments")
        byfile_args_group.add_argument(
            "factories",
            nargs="+",
            metavar="FACTORY_PATH",
            help="path of factory definition files",
        )
        byfile_args_group.add_argument(
            "--regenerate",
            metavar="PROB_REGEN",
            type=probability,
            default=0.0,
            help=(
                "the probability that the factory generation value is not returned as is, but is regenerated. "
                "It affects cases where the original factory returns a value that is not completely random."
            ),
        )
        byfile_args_group.add_argument(
            "--discard",
            metavar="PROB_DISCARD",
            type=probability,
            default=0.0,
            help=(
                "the probability that the factory generation value is not returned as is, but is discarded. "
                "If discarded, the number of times the value is generated is less than "
                "'--repeat'/'-r' or '--list'/'-L' or '--csv'."
            ),
        )
        byfile_args_group.add_argument(
            "--csv",
            metavar="ROW_NUM",
            type=positive_int,
            help="if specified, it outputs generated ROW_NUM objects as CSV. "
            "When using this option, it is recommended to use a factory that generates dictionaries and "
            "to define CSV_COLUMNS in the definition file to specify the fields of the CSV.",
        )
        byfile_args_group.add_argument(
            "--error-on-factory-stopped",
            action="store_true",
            help="If specified, error is occurred in case the factory cannot generate value due to StopIteration. "
            "If not specified, the generation simply stops in the case.",
        )
        add_common_arguments(byfile_parser)

        return byfile_parser

    def validate_parser(self, args: Args, subparser: argparse.ArgumentParser):
        if args.output_fmt == "repr" and args.csv is not None:
            subparser.error("argument --csv: not allowed with argument --repr")
        elif args.output_fmt == "json" and args.csv is not None:
            subparser.error("argument --csv: not allowed with argument --json")
        elif args.list is not None and args.csv is not None:
            subparser.error("argument --csv: not allowed with argument --list/-L")

        if args.regenerate > REGENERATE_PROB_MAX:
            prob_max = Fraction(REGENERATE_PROB_MAX)
            subparser.error(
                f"argument --regenerate: must be lower than or equal to {prob_max}"
            )

    def build_args(
        self, args: Args
    ) -> t.Tuple[t.Sequence[t.Any], t.Mapping[str, t.Any]]:
        pass

    def get_factory_constructor(self) -> t.Callable:
        pass



import random
from abc import ABC, abstractmethod
import typing as t
from fractions import Fraction
from random import Random

from .._utils.nullsafe import dfor

T = t.TypeVar("T")
R = t.TypeVar("R")

REGENERATE_PROB_MAX = float(Fraction(1023, 1024))


class Factory(ABC, t.Generic[T]):
    def next(
        self,
        *,
        raise_on_factory_stopped: bool = False,
    ) -> T:
        """Generate a value randomly according to the rules specified when assembling the factory.

        Parameters
        ----------
        raise_on_factory_stopped : bool, default=False
            If True, raises `FactoryStopException` in case the factory cannot generate value due to `StopIteration`.
            If False, simply raises `StopIteration`.

        Returns
        -------
        T
            a value generated randomly
        """
        try:
            return self._next()
        except StopIteration:
            if raise_on_factory_stopped:
                raise FactoryStopException()
            else:
                raise

    @abstractmethod
    def _next(self) -> T:
        pass

    def or_none(
        self,
        prob: float = 0.1,
        *,
        lazy_choice: bool = False,
        rnd: t.Optional[Random] = None,
    ) -> "Factory[t.Union[T, None]]":
        """Returns a factory whose result may be None with the specified probability.

        Examples
        --------
        >>> import randog
        >>>
        >>> factory = randog.factory.randstr().or_none(0.2)
        >>>
        >>> generated = factory.next()
        >>> assert generated is None or isinstance(generated, str)

        Parameters
        ----------
        prob : float, default=0.1
            Probability that the result is None
        lazy_choice : bool, optional
            If it is True, when generating a value,
            first generate value with the base factory and then decides whether to adopt it or None.
            Otherwise, it first decides whether to return None or generate a value and return it,
            and then generates a value only if it is returned.
        rnd : Random, optional
            random number generator to be used

        Returns
        -------
        Factory[T|None]
            A factory whose result may be None with the specified probability.
        """
        import randog.factory

        return randog.factory.union(
            self,
            randog.factory.const(None),
            weights=[1 - prob, prob],
            lazy_choice=lazy_choice,
            rnd=rnd,
        )

    def post_process(self, post_process: t.Callable[[T], R]) -> "Factory[R]":
        """Returns a factory whose result will be modified by `post_process`

        Examples
        --------
        >>> import randog
        >>>
        >>> # use post_process to format the random decimal value
        >>> factory = (
        ...     randog.factory.randdecimal(0, 50000, decimal_len=2)
        ...                  .post_process(lambda x: f"${x:,}")
        ... )
        >>>
        >>> # examples: '$12,345.67', '$3,153.21', '$12.90', etc.
        >>> generated = factory.next()
        >>> assert isinstance(generated, str)
        >>> assert generated[0] == "$"

        Parameters
        ----------
        post_process : Callable[[T], R]
            the mapping to modify the result

        Returns
        -------
        Factory[R]
            A factory whose result will be modified by `post_process`.
        """
        return PostFactory(self, post_process)

    def iter(
        self,
        size: int,
        *,
        regenerate: float = 0.0,
        discard: float = 0.0,
        raise_on_factory_stopped: bool = False,
        rnd: t.Optional[Random] = None,
    ) -> t.Iterator[T]:
        """Returns an iterator which serves result randomly `size` times.

        Examples
        --------
        >>> import randog
        >>> factory = randog.factory.randstr(length=5)
        >>>
        >>> for result in factory.iter(10):
        ...     assert isinstance(result, str)
        >>>
        >>> results = list(factory.iter(5))
        >>> assert len(results) == 5

        Parameters
        ----------
        size : int
            the number of the iterator.
            However, if the argument `raise_on_factory_stopped` is not True,
            fewer iterations than the specified `size` will be executed if the factory is stopped.
            Also, if the argument `discard` is specified, the size may be less.
        regenerate : float, default=0.0
            the probability that the original factory generation value is not returned as is, but is regenerated.
            It affects cases where the original factory returns a value that is not completely random.
        discard : float, default=0.0
            the probability that the original factory generation value is not returned as is, but is discarded.
            If discarded, the number of times the value is generated is less than `size`.
        raise_on_factory_stopped : bool, default=False
            If True, the iteration raises `FactoryStopException` in case the factory cannot generate value due to
            `StopIteration`.
            If False, the iteration simply stops in the case.
        rnd : Random, optional
            random number generator to be used

        Returns
        -------
        Iterator[T]
            An iterator
        """
        return FactoryIter(
            self,
            size,
            regenerate=regenerate,
            discard=discard,
            raise_on_factory_stopped=raise_on_factory_stopped,
            rnd=rnd,
        )

    def infinity_iter(
        self,
        *,
        regenerate: float = 0.0,
        raise_on_factory_stopped: bool = False,
        rnd: t.Optional[Random] = None,
    ) -> t.Iterator[T]:
        """Returns an infinity iterator which serves result randomly.

        The result is INFINITY so do NOT use it directly with `for`, `list`, and so on.

        However, if the argument `raise_on_factory_stopped` is not True,
        the iterator will be stopped if the factory is stopped.

        Examples
        --------
        >>> import randog
        >>> factory = randog.factory.randstr(length=5)
        >>>
        >>> keys = ["foo", "bar"]
        >>> for k, v in zip(keys, factory.infinity_iter()):
        ...     assert k in keys
        ...     assert isinstance(v, str)

        Parameters
        ----------
        regenerate : float, default=0.0
            the probability that the original factory generation value is not returned as is, but is regenerated.
            It affects cases where the original factory returns a value that is not completely random.
        raise_on_factory_stopped : bool, default=False
            If True, the iteration raises `FactoryStopException` in case the factory cannot generate value due to
            `StopIteration`.
            If False, the iteration simply stops in the case.
        rnd : Random, optional
            random number generator to be used

        Returns
        -------
        Iterator[T]
            An infinity iterator
        """
        return FactoryIter(
            self,
            None,
            regenerate=regenerate,
            raise_on_factory_stopped=raise_on_factory_stopped,
            rnd=rnd,
        )


class PostFactory(Factory[R], t.Generic[T, R]):
    _base_factory: Factory[T]
    _post_process: t.Callable[[T], R]

    def __init__(self, base_factory: Factory[T], post_process: t.Callable[[T], R]):
        self._base_factory = base_factory
        self._post_process = post_process

    def _next(self) -> R:
        pre_result = self._base_factory.next()
        return self._post_process(pre_result)


class FactoryIter(t.Generic[T], t.Iterator[T]):
    _rnd: random.Random
    _factory: Factory[T]
    _size: t.Union[int, float]
    _regenerate_prob: float
    _discard_prob: float
    _raise_on_factory_stopped: bool
    _count: int = 0

    def __init__(
        self,
        factory: Factory[T],
        size: t.Union[int, None],
        *,
        regenerate: float = 0.0,
        discard: float = 0.0,
        raise_on_factory_stopped: bool = False,
        rnd: t.Optional[Random] = None,
    ):
        self._rnd = dfor(rnd, random.Random())
        self._factory = factory

        if regenerate > 0 and discard > 0:
            raise ValueError(
                "`regenerate` and `discard` cannot be specified at the same time"
            )
        if regenerate < 0 or REGENERATE_PROB_MAX < regenerate:
            raise ValueError(
                "the probability `regenerate` must range from 0 to 1023/1024"
            )
        if discard < 0 or 1 < discard:
            raise ValueError("the probability `discard` must range from 0 to 1")

        self._size = size if size is not None else float("Infinity")
        self._regenerate_prob = regenerate
        self._discard_prob = discard
        self._raise_on_factory_stopped = raise_on_factory_stopped

    def __next__(self) -> T:
        while True:
            if self._count >= self._size:
                raise StopIteration()

            self._count += 1

            # Regenerate until break
            while True:
                generated = self._factory.next(
                    raise_on_factory_stopped=self._raise_on_factory_stopped
                )

                if (
                    self._regenerate_prob <= 0
                    or self._rnd.random() >= self._regenerate_prob
                ):
                    break

            # Probabilistically, discard the value generated this time and proceed to the next one
            if self._discard_prob > 0 and self._rnd.random() < self._discard_prob:
                continue

            return generated


class FactoryStopException(Exception):
    def __init__(self):
        super().__init__("the factory stopped generating")



import typing as t
from random import Random

from ._base import Factory
from .._utils.nullsafe import dfor
from ..exceptions import FactoryConstructionError

T = t.TypeVar("T", bound=t.Sequence)


def randlist(
    *items: Factory,
    length: t.Union[int, Factory[int], None] = None,
    type: t.Callable[[t.Iterator[t.Any]], T] = list,
    rnd: t.Optional[Random] = None,
    items_list: t.Optional[t.Sequence[Factory]] = None,
) -> Factory[list]:
    """Return a factory generating random list.

    Parameters
    ----------
    items : Factory
        the factories of each item. If `items_list` is specified, `items` will be ignored.
    length : int|Factory[int], optional
        length of generated list.
        If not specified, the length of generated list will be equals to the number of `items`.
    type : type, default=list
        the type of generated object
    rnd : Random, optional
        random number generator to be used
    items_list : Sequence[Factory], optional
        the factories of each item. Use when positional arguments cannot be specified.

    Raises
    ------
    FactoryConstructionError
        When the specified generating conditions are inconsistent.
    """
    if items_list is not None:
        items = items_list

    return ListRandomFactory(items, length=length, type=type, rnd=rnd)


class ListRandomFactory(Factory[list], t.Generic[T]):
    """factory generating random list values"""

    _random: Random
    _length: t.Union[int, Factory[int]]
    _factories: t.Sequence[Factory]
    _type: t.Callable[[t.Iterator[t.Any]], T]

    def __init__(
        self,
        items: t.Sequence[Factory],
        *,
        length: t.Union[int, Factory[int], None] = None,
        type: t.Callable[[t.Iterator[t.Any]], T] = list,
        rnd: t.Optional[Random] = None,
    ):
        """Return a factory generating random list values.

        Parameters
        ----------
        items : t.Sequence[Factory]
            the factories of each item
        length : int|Factory[int], optional
            length of generated list.
            If not specified, the length of generated list will be equals to the number of `items`.
        type : type, default=list
            the type of generated object
        rnd : Random, optional
            random number generator to be used

        Raises
        ------
        FactoryConstructionError
            When the specified generating conditions are inconsistent.
        """
        self._random = dfor(rnd, Random())
        self._length = dfor(length, len(items))
        self._factories = items
        self._type = type

        non_factory_item_indexes = [
            str(i)
            for i, factory in enumerate(self._factories)
            if not isinstance(factory, Factory)
        ]
        if len(non_factory_item_indexes) > 0:
            raise FactoryConstructionError(
                f"randlist received non-factory object for item: index {', '.join(non_factory_item_indexes)}"
            )

        if isinstance(self._length, Factory) and len(self._factories) == 0:
            raise FactoryConstructionError(
                "the factory of element must be given to randlist() if length is at random"
            )
        if (
            not isinstance(self._length, Factory)
            and self._length > 0
            and len(self._factories) == 0
        ):
            raise FactoryConstructionError(
                "the factory of element must be given to randlist() if length is positive"
            )

    def _factory_generator(self, length) -> t.Iterator[Factory]:
        generated = 0
        last_factory = None
        for factory in self._factories:
            if generated >= length:
                return
            yield factory
            last_factory = factory
            generated += 1
        while generated < length:
            yield last_factory
            generated += 1

    def _next_generator(self) -> t.Iterator[t.Any]:
        length = self._next_length()
        return _GeneratedAsIter(self._factory_generator(length))

    def _next_length(self) -> int:
        if isinstance(self._length, Factory):
            return self._length.next()
        else:
            return self._length

    def _next(self) -> T:
        try:
            return self._type(self._next_generator())
        except _StopIterationOfItem:
            raise StopIteration()


class _GeneratedAsIter(t.Iterator[t.Any]):
    _factories_iter: t.Iterator[Factory]

    def __init__(self, factories_iter: t.Iterator[Factory]):
        self._factories_iter = factories_iter

    def __next__(self) -> t.Any:
        factory = next(self._factories_iter)
        try:
            return factory.next()
        except StopIteration:
            raise _StopIterationOfItem()


class _StopIterationOfItem(Exception):
    pass



import datetime as dt
from decimal import Decimal
import re
import typing as t


class TimedeltaTuple(t.NamedTuple):
    days: int = 0
    hours: int = 0
    minutes: int = 0
    seconds: int = 0
    milliseconds: int = 0
    microseconds: int = 0

    @classmethod
    def of(cls, value: dt.timedelta) -> "TimedeltaTuple":
        minutes, seconds = divmod(value.seconds, 60)
        hours, minutes = divmod(minutes, 60)

        milliseconds, microseconds = divmod(value.microseconds, 1000)

        return TimedeltaTuple(
            days=value.days,
            hours=hours,
            minutes=minutes,
            seconds=seconds,
            milliseconds=milliseconds,
            microseconds=microseconds,
        )


def to_iso(
    value: dt.timedelta,
    *,
    exclude_milliseconds: bool = False,
    point_char: str = ".",
) -> str:
    """Convert timedelta into ISO-8601 format

    Because the timedelta object manages the date part in terms of days,
    the date parts in the output string are also expressed in terms of days only.
    For example, 365 days are represented as "P365D" instead of "P1Y".

    Parameters
    ----------
    value : timedelta
        the timedelta object
    exclude_milliseconds : bool
        If it is True, the microsecond part is excluded.
    point_char : str
        the point character. For example, if "," is specified, 500 milliseconds would be "PT0,5S".

    Returns
    -------
    str
        the ISO-8601 format of the received timedelta
    """
    if value < dt.timedelta():
        result = "-P"
        value = -1 * value
    else:
        result = "P"

    days, hours, minutes, seconds, _, _ = TimedeltaTuple.of(value)

    # add decimal part into seconds
    if not exclude_milliseconds and value.microseconds != 0:
        seconds += Decimal(value.microseconds).scaleb(-6).normalize()

    if value.days > 0:
        result += f"{days}D"

    result_time_part = ""
    if hours != 0:
        result_time_part += f"{hours}H"
    if minutes != 0:
        result_time_part += f"{minutes}M"
    if seconds != 0:
        result_time_part += f"{seconds}S".replace(".", point_char)

    if len(result_time_part) > 0:
        result += "T" + result_time_part

    # "P" is invalid even if the value equals 0
    if len(result) <= 1:
        return "PT0S"

    return result


FROM_ISO_REGEX = re.compile(
    r"^(-?)P(?:(-?\d+(?:\.\d+)?)D)?(?:T(?:(-?\d+(?:\.\d+)?)H)?(?:(-?\d+(?:\.\d+)?)M)?(?:(-?\d+(?:\.\d+)?)S)?)?$"
)


@t.overload
def from_iso(
    value: str, *, returns_none_by_fmt_error: t.Literal[False] = False
) -> dt.timedelta:
    pass


@t.overload
def from_iso(
    value: str, *, returns_none_by_fmt_error: t.Literal[True]
) -> t.Optional[dt.timedelta]:
    pass


def from_iso(
    value: str, *, returns_none_by_fmt_error: bool = False
) -> t.Optional[dt.timedelta]:
    """Converts a string of the ISO-8601 format to timedelta.

    Examples
    --------
    >>> from datetime import timedelta
    >>> import randog.timedelta_util as timedelta_util
    >>>
    >>> assert timedelta_util.from_iso("P1D") == timedelta(days=1)
    >>> assert timedelta_util.from_iso("PT3H30M") == timedelta(hours=3, minutes=30)

    Parameters
    ----------
    value : str
        the string of the ISO-8601 format
    returns_none_by_fmt_error : bool
        if True is specified, when the specified string is illegal, returns None instead of raising an exception.

    Returns
    -------
    timedelta
        the timedelta object for the received value
    """
    match = re.match(FROM_ISO_REGEX, value)

    if not match:
        if returns_none_by_fmt_error:
            return None
        else:
            raise ValueError(f"Invalid isoformat string: {value}")

    sign = -1 if match.group(1) == "-" else 1
    days, hours, minutes, seconds = map(_float_or_zero, match.groups()[1:])

    return sign * dt.timedelta(
        days=days,
        hours=hours,
        minutes=minutes,
        seconds=seconds,
    )


def to_str(
    value: dt.timedelta,
) -> str:
    """Convert timedelta into the simple format

    Parameters
    ----------
    value : timedelta
        the timedelta object

    Returns
    -------
    str
        the simple format of the received timedelta
    """
    if value < dt.timedelta():
        result = "-"
        value = -1 * value
    else:
        result = ""

    value_tuple = TimedeltaTuple.of(value)

    if value_tuple.days > 0:
        result += f"{value_tuple.days}d"
    if value_tuple.hours != 0:
        result += f"{value_tuple.hours}h"
    if value_tuple.minutes != 0:
        result += f"{value_tuple.minutes}m"
    if value_tuple.seconds != 0:
        result += f"{value_tuple.seconds}s"
    if value_tuple.milliseconds != 0:
        result += f"{value_tuple.milliseconds}ms"
    if value_tuple.microseconds != 0:
        result += f"{value_tuple.microseconds}us"

    # "P" is invalid even if the value equals 0
    if len(result) == 0:
        return "0s"

    return result


FROM_STR__VALIDATE__REGEXP = re.compile(r"^([-+]?(\d+[a-zA-Z]+|0))+$")
FROM_STR__SPLIT_TERM__REGEXP = re.compile(r"(?:^|[-+])(?:\d+[a-zA-Z]+)+")


def from_str(value: str) -> dt.timedelta:
    """Converts a string of the simple format to timedelta.

    Examples
    --------
    >>> from datetime import timedelta
    >>> import randog.timedelta_util as timedelta_util
    >>>
    >>> assert timedelta_util.from_str("1h30m") == timedelta(hours=1, minutes=30)
    >>> assert timedelta_util.from_str("1d2h3m4s5ms6us") == timedelta(
    ...             days=1, hours=2, minutes=3, seconds=4, milliseconds=5, microseconds=6)

    Parameters
    ----------
    value : str
        the string of the format

    Returns
    -------
    timedelta
        the timedelta object for the received value
    """
    if not re.match(FROM_STR__VALIDATE__REGEXP, value):
        raise TimedeltaExpressionError(f"illegal timedelta expression: {value}")

    result = dt.timedelta(0)

    try:
        for combined_term in re.finditer(FROM_STR__SPLIT_TERM__REGEXP, value):
            result += _term_from_str(combined_term.group(0))
    except TimedeltaExpressionError as e:
        raise TimedeltaExpressionError(f"illegal timedelta expression: {value}") from e

    return result


__TO_FMT_MAP: t.Mapping[str, t.Callable[[dt.timedelta], str]] = {
    "%D": lambda td: str(td.days),
    "%H": lambda td: f"{td.seconds // 3600:02}",
    "%tH": lambda td: f"{td.total_seconds() // 3600:.0f}",
    "%M": lambda td: f"{(td.seconds % 3600) // 60:02}",
    "%tM": lambda td: f"{td.total_seconds() // 60:.0f}",
    "%S": lambda td: f"{td.seconds % 60:02}",
    "%tS": lambda td: f"{int(td.total_seconds())}",
    "%f": lambda td: f"{td.microseconds:06}",
}


def to_fmt(value: dt.timedelta, fmt: str) -> str:
    """Convert timedelta into the specified format

    The following is a list of all the format codes:

    * :code:`%D`: replaced by :code:`timedelta.days`
    * :code:`%H`: hours part (zero-padded to 2 digits)
    * :code:`%tH`: total duration in hours (rounded down)
    * :code:`%M`: minutes part (zero-padded to 2 digits)
    * :code:`%tH`: total duration in minutes (rounded down)
    * :code:`%S`: seconds part (zero-padded to 2 digits)
    * :code:`%tS`: total duration in seconds (rounded down)
    * :code:`%f`: decimal part in seconds (zero-padded to 6 digits)

    Parameters
    ----------
    value : timedelta
        the timedelta object
    fmt : str
        the format string

    Returns
    -------
    str
        the specified format of the specified timedelta
    """
    result = ""
    replace_map = {}

    i = 0
    while i < len(fmt):
        if fmt[i] == "%":
            if fmt[i + 1] == "t":
                fmt_exp = fmt[i : i + 3]
            else:
                fmt_exp = fmt[i : i + 2]

            if fmt_exp not in replace_map:
                try:
                    replace_map[fmt_exp] = __TO_FMT_MAP[fmt_exp](value)
                except KeyError:
                    raise ValueError("Invalid format string")
            result += replace_map[fmt_exp]

            i += len(fmt_exp)
        else:
            result += fmt[i]
            i += 1

    return result


TERM_FROM_STR__VALIDATE__REGEXP = re.compile(r"^[-+]?(?:\d+[a-zA-Z]+)+$")
TERM_FROM_STR__SPLIT_SINGLE_UNIT_TERM__REGEXP = re.compile(r"(\d+)([a-zA-Z]+)")


def _term_from_str(term_str: str) -> dt.timedelta:
    if not re.match(TERM_FROM_STR__VALIDATE__REGEXP, term_str):
        raise ValueError(f"illegal timedelta term expression: {term_str}")

    sign = -1 if term_str.startswith("-") else 1
    result = dt.timedelta(0)

    try:
        for single_unit_term in re.finditer(
            TERM_FROM_STR__SPLIT_SINGLE_UNIT_TERM__REGEXP, term_str
        ):
            numeric_str, unit = single_unit_term.group(1, 2)
            numeric = int(numeric_str)
            result += _UNIT_TERM_CONSTRUCTOR[unit](numeric)
    except (TypeError, KeyError):
        raise TimedeltaExpressionError(f"illegal timedelta term expression: {term_str}")

    return sign * result


_UNIT_TERM_CONSTRUCTOR = {
    "d": lambda num: dt.timedelta(days=num),
    "h": lambda num: dt.timedelta(hours=num),
    "m": lambda num: dt.timedelta(minutes=num),
    "s": lambda num: dt.timedelta(seconds=num),
    "ms": lambda num: dt.timedelta(milliseconds=num),
    "us": lambda num: dt.timedelta(microseconds=num),
}


def _float_or_zero(value: t.Union[str, None]) -> float:
    if not value:
        return 0
    else:
        return float(value)


class TimedeltaExpressionError(ValueError):
    pass



import enum
import numbers
from random import Random
import typing as t

from ._base import Factory
from .._utils.nullsafe import dfor
from ..exceptions import FactoryConstructionError


def randchoice(
    *values: t.Any,
    weights: t.Optional[t.Sequence[float]] = None,
    rnd: t.Optional[Random] = None,
) -> Factory[t.Any]:
    """Return a factory choosing one of values.

    Parameters
    ----------
    values : Any
        the values
    weights : Sequence[float], optional
        the probabilities that each value is chosen.
        The length must equal to the number of values.
    rnd : Random, optional
        random number generator to be used

    Raises
    ------
    FactoryConstructionError
        No values are specified.
    """
    return ChoiceRandomFactory(values, weights=weights, rnd=rnd)


def randenum(
    enum_cls: t.Type[enum.Enum],
    weights: t.Optional[t.Callable[[t.Any], float]] = None,
    rnd: t.Optional[Random] = None,
) -> Factory[t.Any]:
    """Return a factory choosing one of specified enum.

    Parameters
    ----------
    enum_cls : Type[Enum]
        the enum class
    weights : Callable[[Enum], float], optional
        the probabilities that each value is chosen.
    rnd : Random, optional
        random number generator to be used

    Raises
    ------
    FactoryConstructionError
        No values are specified.
    """
    values = [*enum_cls]
    weights_list = [weights(value) for value in values] if weights is not None else None

    # validate weight_list
    if weights_list is not None:
        for weight in weights_list:
            __validate_num(
                weight,
                err_msg="the weights must serve weight for each enum value",
            )

    return ChoiceRandomFactory(values, weights=weights_list, rnd=rnd)


class ChoiceRandomFactory(Factory[t.Any]):
    """factory choosing one of values."""

    _random: Random
    _weights: t.Optional[t.Sequence[float]]
    _values: t.Sequence[t.Any]

    def __init__(
        self,
        values: t.Sequence,
        weights: t.Optional[t.Sequence[float]] = None,
        rnd: t.Optional[Random] = None,
    ):
        """Return a factory choosing one of values.

        Parameters
        ----------
        values : Any
            the values
        weights : Sequence[float], optional
            the probabilities that each value is chosen.
            The length must equal to the number of values.
        rnd : Random, optional
            random number generator to be used

        Raises
        ------
        FactoryConstructionError
            No values are specified.
        """
        self._random = dfor(rnd, Random())
        self._weights = weights
        self._values = values

        if len(values) <= 0:
            raise FactoryConstructionError("empty candidate for randchoice")
        if self._weights is not None and len(self._weights) != len(self._values):
            raise FactoryConstructionError(
                "the number of weights must match the candidates"
            )

    def _next(self) -> t.Any:
        return self._random.choices(self._values, self._weights)[0]


def __validate_num(value, *, err_msg) -> numbers.Real:
    if isinstance(value, numbers.Real) and not isinstance(value, bool):
        return value
    else:
        raise FactoryConstructionError(err_msg)



import argparse
import typing as t

import randog.factory
from ..._utils.type import probability
from .. import Args, Subcmd
from ._base import SubcmdDef, add_common_arguments


class SubcmdDefBool(SubcmdDef):
    def cmd(self) -> Subcmd:
        return Subcmd.Bool

    def add_parser(self, subparsers) -> argparse.ArgumentParser:
        bool_parser = subparsers.add_parser(
            Subcmd.Bool.value,
            usage="python -m randog bool [PROP_TRUE] [common-options]",
            description="It generates boolean values.",
            add_help=False,
        )
        bool_args_group = bool_parser.add_argument_group("arguments")
        bool_args_group.add_argument(
            "prop_true",
            type=probability,
            default=0.5,
            nargs="?",
            metavar="PROP_TRUE",
            help="the probability of True",
        )
        add_common_arguments(bool_parser)

        return bool_parser

    def validate_parser(self, args: Args, subparser: argparse.ArgumentParser):
        pass

    def build_args(
        self, args: Args
    ) -> t.Tuple[t.Sequence[t.Any], t.Mapping[str, t.Any]]:
        return (args.get("prop_true"),), {}

    def get_factory_constructor(self) -> t.Callable:
        return randog.factory.randbool



import typing as t

T = t.TypeVar("T")
R = t.TypeVar("R")


def dfor(value: t.Optional[T], default: T) -> T:
    if value is None:
        return default
    else:
        return value


def dforc(m: t.Callable[[T], R], value: t.Optional[T]) -> t.Optional[R]:
    if value is None:
        return None
    else:
        return m(value)



import datetime
from decimal import Decimal
import typing as t
from random import Random

import sqlalchemy
import sqlalchemy.orm
import sqlalchemy.sql.schema

import randog.factory
from .._utils.nullsafe import dfor
from ..factory import Factory, randdict, from_example
from ..exceptions import FactoryConstructionError

M = t.TypeVar("M")


@t.overload
def factory(
    model: t.Type[M],
    override_columns: t.Mapping[str, t.Union[Factory, t.Any]] = None,
    *,
    as_dict: t.Literal[False] = False,
    rnd: t.Optional[Random] = None,
) -> Factory[M]:
    ...


@t.overload
def factory(
    model: t.Type[M],
    override_columns: t.Mapping[str, t.Union[Factory, t.Any]] = None,
    *,
    as_dict: t.Literal[True],
    rnd: t.Optional[Random] = None,
) -> Factory[dict]:
    ...


def factory(
    model: t.Type[M],
    override_columns: t.Mapping[str, t.Union[Factory, t.Any]] = None,
    *,
    as_dict: bool = False,
    rnd: t.Optional[Random] = None,
) -> t.Union[Factory[M], Factory[dict]]:
    """Return a factory generating Model instances randomly.

    Parameters
    ----------
    model : Type[Model]
        the model class of sqlalchemy
    override_columns : Mapping[str, Union[Factory, Column, Any]], optional
        For each column that is specified in this dictionary,
        create a factory using it instead of the column of the model class.
        If factory is specified as value of this dict, it is used to generate the field;
        if other is specified,
        the factory obtained by using `from_example` with it as an example is used to generate the field.
    as_dict : bool, default=False
        If it is True, values generated by the factory will be dict, not model instances.
    rnd : Random, optional
        random number generator to be used
    """
    if override_columns is None:
        override_columns = dict()

    mapper = sqlalchemy.inspect(model, raiseerr=False)
    if not isinstance(mapper, sqlalchemy.orm.Mapper):
        raise FactoryConstructionError("specified object is not model of sqlalchemy")

    factories_of_column = {
        key: _factory_from_column_etc(override_columns.get(key, column), rnd=rnd)
        for key, column in mapper.columns.items()
    }

    dict_factory = randdict(factories_of_column, rnd=rnd)

    if as_dict:
        return dict_factory
    else:
        return dict_factory.post_process(lambda r: model(**r))


def _factory_from_column_etc(
    column: t.Any,
    *,
    rnd: t.Optional[Random] = None,
) -> Factory:
    if isinstance(column, Factory):
        return column
    elif isinstance(column, sqlalchemy.Column):
        return factory_from_column(column, rnd=rnd)
    else:
        return from_example(column, rnd=rnd)


def factory_from_column(
    column: sqlalchemy.sql.schema.Column,
    *,
    rnd: t.Optional[Random] = None,
) -> Factory:
    """Return a factory generating values that matches the column.

    Parameters
    ----------
    column : Column
        the column of sqlalchemy
    rnd : Random, optional
        random number generator to be used
    """

    # type checks of `column`
    column_type, column_python_type = _get_column_types(column)
    enums = getattr(column_type, "enums", None) if column_type is not None else None
    autoincrement = getattr(column, "autoincrement", False)

    # normalize example for the function `from_example`.
    # If this is difficult, create the factory directly.
    if enums is not None and column_python_type == str:
        # use randchoice instead from_example(str) if it is enumeration
        factory_ = randog.factory.randchoice(*enums, rnd=rnd)
    elif autoincrement and column_python_type == int:
        # use increment instead from_example(int) if it is auto incremental
        factory_ = randog.factory.increment(maximum=2**31 - 1, rnd=rnd)
    elif column_python_type in __FACTORY_FOR_SPEC_TYPE:
        factory_ = __FACTORY_FOR_SPEC_TYPE[column_python_type](
            column_type,
            rnd=rnd,
        )
    elif isinstance(column_python_type, type):
        factory_ = randog.factory.from_example(column_python_type, rnd=rnd)
    else:
        raise FactoryConstructionError(
            f"cannot create factory from sqlalchemy.Column type of {column_type}: not implemented"
        )

    # Create a factory based on the normalized example.
    if not isinstance(factory_, Factory):
        factory_ = randog.factory.from_example(factory_, rnd=rnd)
    # make the factory nullable
    if hasattr(column, "nullable") and column.nullable:
        factory_ = factory_.or_none(rnd=rnd)

    return factory_


def _factory_for_str_column(
    column_type,
    rnd: t.Optional[Random] = None,
):
    length = getattr(column_type, "length", None)
    if length is not None:
        return randog.factory.randstr(length=length, rnd=rnd)
    else:
        return str


def _factory_for_numeric_column(
    column_type,
    rnd: t.Optional[Random] = None,
) -> Factory:
    precision = getattr(column_type, "precision", None)
    scale = getattr(column_type, "scale", None)

    if precision is not None:
        scale_n = dfor(scale, 0)

        maximum = Decimal("1").scaleb(precision - scale_n) - Decimal("1").scaleb(
            -scale_n
        )
        minimum = -maximum
    else:
        minimum = None
        maximum = None

    return randog.factory.randdecimal(minimum, maximum, decimal_len=scale, rnd=rnd)


def _factory_for_datetime_column(
    column_type,
    rnd: t.Optional[Random] = None,
) -> Factory[datetime.datetime]:
    use_timezone = getattr(column_type, "timezone", None)

    if use_timezone:
        return randog.factory.randdatetime(tzinfo=datetime.timezone.utc, rnd=rnd)
    else:
        return randog.factory.randdatetime(tzinfo=None, rnd=rnd)


def _factory_for_time_column(
    column_type,
    rnd: t.Optional[Random] = None,
) -> Factory[datetime.time]:
    use_timezone = getattr(column_type, "timezone", None)

    if use_timezone:
        return randog.factory.randtime(tzinfo=datetime.timezone.utc, rnd=rnd)
    else:
        return randog.factory.randtime(tzinfo=None, rnd=rnd)


__FACTORY_FOR_SPEC_TYPE = {
    str: _factory_for_str_column,
    Decimal: _factory_for_numeric_column,
    datetime.datetime: _factory_for_datetime_column,
    datetime.time: _factory_for_time_column,
}


def _get_column_types(column) -> t.Tuple[t.Any, t.Any]:
    # type checks of `column`
    if hasattr(column, "type"):
        column_type = column.type
    else:
        raise FactoryConstructionError(
            f"cannot create factory from {column}: not implemented"
        )
    if hasattr(column_type, "python_type"):
        column_python_type = column_type.python_type
    else:
        raise FactoryConstructionError(
            f"cannot create factory from {column}: not implemented"
        )

    return column_type, column_python_type



class FactoryConstructionError(Exception):
    _message: str

    @property
    def message(self) -> str:
        return self._message

    def __init__(self, message: str):
        super().__init__(message)
        self._message = message



"""
the package contains the Args of module execution and its builder
"""

import argparse
import typing as t

from ._subcmd import Subcmd


class Args:
    _args: argparse.Namespace

    def __init__(self, argv: t.Sequence[str]):
        from ._subcmd_def import iter_subcmd_def

        parser = argparse.ArgumentParser(
            prog="randog",
            description="It generates values randomly according to the specified mode and arguments.",
        )
        subparsers = parser.add_subparsers(
            required=True,
            dest="sub",
            metavar="MODE",
            help=f"mode of value generation; candidates: {', '.join(map(lambda c: c.value, Subcmd))}. "
            "For more information, see the command 'randog MODE --help'.",
        )

        # create subcommands and add them to the ArgumentParser
        subcmd_parsers = {}
        for subcmd in iter_subcmd_def():
            subcmd_parsers[subcmd.cmd()] = subcmd.add_parser(subparsers)

        # parse the arguments
        self._args = parser.parse_args(argv[1:])

        # validate arguments for subcommands
        for subcmd in iter_subcmd_def():
            subcmd.validate_parser(self, subcmd_parsers[subcmd.cmd()])

    @property
    def sub_cmd(self) -> Subcmd:
        return Subcmd(self._args.sub)

    @property
    def factories(self) -> t.Sequence[str]:
        return self._args.factories

    @property
    def repeat(self) -> int:
        return self._args.repeat

    @property
    def list(self) -> t.Optional[int]:
        return self._args.list

    @property
    def output_fmt(self) -> t.Optional[str]:
        return self._args.output_fmt

    @property
    def output_path(self) -> t.Optional[str]:
        return self._args.output

    @property
    def multiple_output_path(self) -> bool:
        if self._args.output is None:
            return False
        else:
            return self.output_path_for(1) != self.output_path_for(2)

    @property
    def iso(self) -> bool:
        if hasattr(self._args, "iso"):
            return self._args.iso
        else:
            return False

    @property
    def format(self) -> t.Optional[str]:
        if hasattr(self._args, "format"):
            return self._args.format
        else:
            return None

    @property
    def regenerate(self) -> float:
        if hasattr(self._args, "regenerate"):
            return self._args.regenerate
        else:
            return 0.0

    @property
    def discard(self) -> float:
        if hasattr(self._args, "discard"):
            return self._args.discard
        else:
            return 0.0

    @property
    def csv(self) -> t.Optional[int]:
        if hasattr(self._args, "csv"):
            return self._args.csv
        else:
            return None

    @property
    def error_on_factory_stopped(self) -> bool:
        return self.get("error_on_factory_stopped", False)

    def output_path_for(self, number: int) -> t.Optional[str]:
        if self._args.output is None:
            return None
        else:
            return self._args.output.format(number)

    def get(self, key: str, default: t.Any = None) -> t.Any:
        if hasattr(self._args, key):
            return getattr(self._args, key)
        else:
            return default



import typing as t
from dataclasses import dataclass


class Example(t.Sequence):
    _objs: t.Sequence

    @t.overload
    def __getitem__(self, index: int) -> t.Any:
        ...

    @t.overload
    def __getitem__(self, index: slice) -> t.Sequence[t.Any]:
        ...

    def __getitem__(self, index: t.Union[int, slice]):
        return self._objs[index]

    def __len__(self) -> int:
        return len(self._objs)

    def __init__(self, *objs):
        self._objs = objs


@dataclass
class DictItemExample:
    example: t.Any
    prop_exists: float = 1.0



import argparse
import typing as t

import randog.factory
from ..._utils.type import non_negative_int, probability
from .. import Args, Subcmd
from ._base import SubcmdDef, add_common_arguments


class SubcmdDefDecimal(SubcmdDef):
    def cmd(self) -> Subcmd:
        return Subcmd.Decimal

    def add_parser(self, subparsers) -> argparse.ArgumentParser:
        decimal_parser = subparsers.add_parser(
            Subcmd.Decimal.value,
            usage="python -m randog decimal [MINIMUM MAXIMUM] [--decimal-len DECIMAL_LENGTH] "
            "[--p-inf PROB_P_INF] [--n-inf PROB_N_INF] [--nan PROB_NAN] [common-options]",
            description="It generates values of type decimal.Decimal.",
            add_help=False,
        )
        decimal_args_group = decimal_parser.add_argument_group("arguments")
        decimal_args_group.add_argument(
            "minimum",
            type=float,
            nargs="?",
            metavar="MINIMUM",
            help="the minimum value. "
            "If not specified, the behavior is left to the specification of randog.factory.randdecimal.",
        )
        decimal_args_group.add_argument(
            "maximum",
            type=float,
            nargs="?",
            metavar="MAXIMUM",
            help="the maximum value. "
            "If not specified, the behavior is left to the specification of randog.factory.randdecimal.",
        )
        decimal_args_group.add_argument(
            "--decimal-len",
            type=non_negative_int,
            default=None,
            metavar="DECIMAL_LENGTH",
            help="the length of decimal part of generated values",
        )
        decimal_args_group.add_argument(
            "--p-inf",
            type=probability,
            default=0.0,
            metavar="PROB_P_INF",
            help="the probability of positive infinity; default=0.0",
        )
        decimal_args_group.add_argument(
            "--n-inf",
            type=probability,
            default=0.0,
            metavar="PROB_N_INF",
            help="the probability of negative infinity; default=0.0",
        )
        decimal_args_group.add_argument(
            "--nan",
            type=probability,
            default=0.0,
            metavar="PROB_NAN",
            help="the probability of NaN; default=0.0",
        )
        add_common_arguments(decimal_parser)

        return decimal_parser

    def validate_parser(self, args: Args, subparser: argparse.ArgumentParser):
        if args.sub_cmd != Subcmd.Decimal:
            return

        iargs, kwargs = self.build_args(args)
        minimum, maximum = iargs
        nan = kwargs["nan"]
        p_inf = kwargs["p_inf"]
        n_inf = kwargs["n_inf"]

        if minimum is not None and maximum is not None and minimum > maximum:
            subparser.error("arguments must satisfy MINIMUM <= MAXIMUM")

        if nan + p_inf + n_inf > 1.0:
            subparser.error(
                "arguments must satisfy that PROB_P_INF + PROB_N_INF + PROB_NAN <= 1.0"
            )

    def build_args(
        self, args: Args
    ) -> t.Tuple[t.Sequence[t.Any], t.Mapping[str, t.Any]]:
        return (args.get("minimum"), args.get("maximum")), {
            "decimal_len": args.get("decimal_len"),
            "p_inf": args.get("p_inf"),
            "n_inf": args.get("n_inf"),
            "nan": args.get("nan"),
        }

    def get_factory_constructor(self) -> t.Callable:
        return randog.factory.randdecimal



from random import Random
import typing as t

from ._base import Factory
from .._utils.nullsafe import dfor
from ..exceptions import FactoryConstructionError


def union(
    *factories: Factory,
    weights: t.Optional[t.Sequence[float]] = None,
    lazy_choice: bool = False,
    rnd: t.Optional[Random] = None,
) -> Factory[t.Any]:
    """Return a factory generating value by one of specified factories.

    Parameters
    ----------
    factories : Factory
        the factories
    weights : Sequence[float], optional
        the probabilities that each factory is chosen.
        The length must equal to the number of factories.
    lazy_choice : bool, optional
        If it is True, when generating a value,
        first generate values with all factories and then decide which of them to adopt.
        Otherwise, it first decides which factory to adopt, and then generates a value using only that factory.
    rnd : Random, optional
        random number generator to be used

    Raises
    ------
    FactoryConstructionError
        No factories are specified.
    """
    if lazy_choice:
        return UnionRandomLazyChoiceFactory(*factories, weights=weights, rnd=rnd)

    else:
        return UnionRandomFactory(*factories, weights=weights, rnd=rnd)


class UnionRandomFactory(Factory[t.Any]):
    """factory generating value by one of specified factories."""

    _random: Random
    _weights: t.Optional[t.Sequence[float]]
    _factories: t.Sequence[Factory]

    def __init__(
        self,
        *factories: Factory,
        weights: t.Optional[t.Sequence[float]] = None,
        rnd: t.Optional[Random] = None,
    ):
        """Return a factory generating value by one of specified factories.

        Parameters
        ----------
        factories : Factory
            the factories
        weights : Sequence[float], optional
            the probabilities that each factory is chosen.
            The length must equal to the number of factories.
        rnd : Random, optional
            random number generator to be used

        Raises
        ------
        FactoryConstructionError
            No factories are specified.
        """
        self._random = dfor(rnd, Random())
        self._weights = weights
        self._factories = factories

        if len(factories) <= 0:
            raise FactoryConstructionError("no factory is given to union()")
        if self._weights is not None and len(self._weights) != len(self._factories):
            raise FactoryConstructionError(
                "the number of weights must match the factories"
            )

    def _next(self) -> t.Any:
        return self._random.choices(self._factories, self._weights)[0].next()


class UnionRandomLazyChoiceFactory(UnionRandomFactory):
    def _next(self) -> t.Any:
        values = [f.next() for f in self._factories]
        return self._random.choices(values, self._weights)[0]
